#
# $Copyright: Copyright (c) 2013 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $RCSfile: MetroMirror.pm,v $ $Revision: 1.2.4.25 $
#

package MetroMirror;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use Date::Parse;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use ag_i18n_inc;

@EXPORT = qw(
);
$VERSION = '1.00';

require 5.000;

# Common variables concerning OS / logical volume management
use POSIX qw(uname);
(my $os_name) = uname;
my $VXDG = "/usr/sbin/vxdg";
my $VXDCTL = "/usr/sbin/vxdctl";
my $VXDISK = "/usr/sbin/vxdisk";

my $is_rac_cluster = 0;

my $sltime = 20;

# Platform specific commands for volume management
my $aix_exportvg = "/usr/sbin/exportvg";
my $aix_importvg = "/usr/sbin/importvg";
my $aix_recreatevg = "/usr/sbin/recreatevg";
my $aix_lspv = "/usr/sbin/lspv";
my $aix_cfgmgr = "/usr/sbin/cfgmgr";
my $aix_fd_vg_ext = "_fd";
my $aix_lv_prefix = "fd_";
my $aix_label_prefix = "/";

# ESX specific variables
my $java_home="/opt/VRTSjre/jre1.5/bin";
my $vcs_home=$VCSHOME;
my $JAVA = "$java_home/java";
my $JAVA_CLASSPATH=join (':', <$VCSHOME/vcsvi-sdk/jars/*.jar>);
my $VMWARE_API_CMD = "$JAVA -cp \"$JAVA_CLASSPATH\" com.vmf.VirtualEntityComDriver";
my $cred_file="/etc/VRTSvcs/conf/config/\.esxlocalhostWSCred";

# Maximum retry count for initial dscli command, in case the HMC has a problem
my $dscli_retry = 2;

%default = {
	resname => '',

	VADCtlMode => '',

	# Element to support single library for different types of
	# replication / fire drill snapshots
	# ========================================================
	rtype => '',

	# Elements to store the resource atributes
	# ========================================

	# dscli contains the information of the following attributes:
	# - DSClihome
	# - HMC1, optional HMC2
	# - User
	# - PasswdFile
	dscli => '',

	# local storage immage ID
	lmigid => '',

	# remote storage image ID
	rmigid => '',

	# vids is a hash with the local volume ID as key as defined by
	# attribute VolIds and remote ID as data retrieved by get_state
	vids => '',

	# vidstate is a hash with the local volume ID as key and local
	# volume state as data
	vidstate => '',

	# DSMon determines if monitor shall connect to the HMC
	DSMon => '',

	# Elements specific for different replication / snapshot modes
	# ============================================================

	# Target resource for MetroMirrorSnap
	tres => '',

	# Target resource attribute list
	targ =>'',

	# Session ID for GlobalMirror
	gmsess => '',

	# Elements holding additional data
	# ================================

	# The OS platform the agent is running
	platform => '',

	# The output of the last dscli command executed
	dscout => '',

	# The overall state of the local array
	state => '',

        # Holds value of ComputeDRSLA attribute
        computedrsla => ''
};

sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my $ret;
	my $self = { %default };

	$self->{'rtype'} = shift;
	$self->{'resname'} = shift;
	$self->{'platform'} = $os_name;

	my @attrs = @_;

	bless($self, $class);

	# Check for VAD controlmode return value if > 0
	my $clus_prod = $ENV{"CLUSTER_PRODUCT"};
	$clus_prod = "VCS" if (! defined($clus_prod));
	$self->{'VADCtlMode'} = 0;
	if ($clus_prod eq "VAD") {
		($ret, $self->{'VADCtlMode'}) = VCSAG_GET_ATTR_VALUE ("ControlMode", -1, 1, @attrs);
		return undef if ($ret != $VCSAG_SUCCESS);
	}
	
	
	if ( $self->{'platform'} =~ /^Windows NT/i) {
		require Win32;
		Win32->import;

		my $vmpath;
		$vmpath = $ENV{"VMPATH"};
		$vmpath = Win32::GetShortPathName($vmpath);
		$VXDCTL = "$vmpath" . "vxdctl";
		$VXDISK = "$vmpath" . "vxdisk";
		$VXDG = "$vmpath" . "vxdg";
	}

	if ($self->{'rtype'} eq 'MMS') {
		return undef if ($self->get_mms_attr(@attrs));
		@attrs = split('\s+', $self->{'targ'});
	} elsif ($self->{'rtype'} eq 'GM') {
		# Global mirror session
		return undef if ($self->get_gm_attr(@attrs));
	}


	return undef if ($self->get_mm_attr(@attrs));

	return $self;
}

sub DESTROY {}

sub Rep_Tech_RPO {
    my $rtype = shift;
    my $ResName = shift;

    unshift(@ARGV,$ResName);

    my $res = MetroMirror->new("GM", @ARGV);
    unless (defined $res) {
        return "Rep_Tech_RPO: Cannot comput RPO, attributes values are invalid";
    }

    if ($res->get_state()) {
        return "Rep_Tech_RPO: Cannot comput RPO, unable to get resource current state";
    }

    if (!($res->{'state'} eq "Copy Pending" || $res->{'state'} eq "Suspended")) {
        VCSAG_LOG_MSG("E", "Cannot comput RPO, source volume state must be either Copy Pending or Suspended. Current state is [$res->{'state'}]", 191, "$res->{'state'}");
        return "Rep_Tech_RPO: Cannot comput RPO, source volume state must be Copy Pending or Suspended";
    }

    my $showgmir = {};
    $showgmir->{'cmd'} = "showgmir";
    push(@{$showgmir->{'flag'}}, "-bnr off", "-v off");

    my $ret = $res->exe_dscli($showgmir);
    if ($ret) {
        return "Rep_Tech_RPO: Cannot comput RPO, dscli command showgmir has failed. Please check engine logs for more details.";
    }

    my $current_time = 0;
    my $cg_time = 0;
    foreach my $line (split('\n', $res->{'dscout'})) {
        if ( $line =~ /Current Time/ ) {
            my (@fields) = split('\s+', $line);
            $current_time = $fields[2]." ".$fields[3];
            VCSAG_LOGDBG_MSG(2, "Current Time = $current_time");
        }
        elsif ( $line =~ /CG Time/ ){
            my (@fields) = split('\s+', $line);
            $cg_time = $fields[2]." ".$fields[3];
            VCSAG_LOGDBG_MSG(2, "CG Time = $cg_time");
        }
    }

    unless ( $current_time && $cg_time ) {
        return "Rep_Tech_RPO: Cannot comput RPO, Unable to get current time and last consistency time value";
    }

    my $after_t = str2time( $current_time );
    my $before_t = str2time( $cg_time );
    my $time_diff = time_format($after_t - $before_t);
    my $out = "Rep_Tech_RPO: Current replication lag (Current Time - CG Time) is $time_diff [hh:mm:ss] as reported by IBM Global Mirror";
        return $out;
}

sub time_format {
    my $time = shift;
    my ($hour, $min, $sec,$t1);
    $hour = "0";
    $min = "0";
    $sec = "0";
    if ($time >= 60){
        $t1 = int ($time / 60);
        $sec = ($time % 60);
        if ($t1 >= 60){
            $hour = int ($t1/60);
            $min = $t1 % 60;
        }else{
            $min = ($t1);
        }
    }
    else {
        $sec = $time;
    }
    return sprintf("%02d:%02d:%02d",$hour,$min,$sec);
}

#
# functions used by more than one copy service
#
sub get_state {
	my @pairs = ("\n");
	my $self = shift;

	$self->{'state'} = "";

	my $lspprc = {};
	$lspprc->{'cmd'} = "lspprc";
	push(@{$lspprc->{'flag'}}, "-fmt delim", "-delim ','");
	push(@{$lspprc->{'flag'}}, "-hdr off", "-bnr off -v off");

	my $con_try = 0;
	while ($self->exe_dscli($lspprc) > 0) {
		if ($con_try >= $dscli_retry) {
			VCSAG_LOG_MSG("E", "Array status request failed with message: $self->{'dscout'}", 111, "$self->{'dscout'}");
			return 1;
		}
		$con_try++;
	}

	foreach my $line (split('\n', $self->{'dscout'})) {
		push(@pairs, $line);
		my ($id, $state, $reason, $type, $slss, undef) = split(/,/, $line);
		if ($self->{'rtype'} =~ /MMS?/) {
			next unless ($type eq "Metro Mirror");
		} elsif ($self->{'rtype'} =~ /GMS?/) {
			next unless ($type eq "Global Copy");
		} else {
			next;
		}
		my ($src_vid, $tgt_vid) = split(/:/, $id);

		# set remote metromirror volID if not already set.
		if (exists $self->{'vids'}->{"$src_vid"}  && ($state !~ /TARGET/i)) {
			$self->{'vids'}->{"$src_vid"} = $tgt_vid;
			$self->{'slss'}->{"$src_vid"} = $slss;
			$self->{'vids_state'}->{"$src_vid"} = $state;
		} elsif (exists $self->{'vids'}->{"$tgt_vid"}  && ($state =~ /TARGET/i)) {
			$self->{'vids'}->{"$tgt_vid"} = $src_vid;
			$self->{'slss'}->{"$tgt_vid"} = $slss;
			$self->{'vids_state'}->{"$tgt_vid"} = $state;
		} else {
			next;
		}

		if ($self->{'state'} eq "") {
			$self->{'state'} = $state;
		} elsif ($state ne $self->{'state'}) {
			# volumes are in an invalid mix of states
			VCSAG_LOG_MSG("E", "Volumes are in an inconsistent mix of states: @pairs", 112, "@pairs");
			$self->{'state'} = "Invalid";
			close(fh);
			return 1;
		}
	}
	close(fh);

	# Check if all volumes are mapped
	my $no_match = 0;
	foreach my $vkey (keys %{$self->{'vids'}}) {
		if ($self->{'vids'}->{"$vkey"} eq "") {
			VCSAG_LOG_MSG("E", "Volume $vkey has no matching replicated target volume", 113, "$vkey");
			$no_match = 1;
		}
	}

	return $no_match;
}

sub failover_required {
	my ($self) = shift;

	if ($self->{'rtype'} eq "MM") {
		return 0 if ($self->{'state'} eq "Full Duplex");
		return 0 if ($self->{'state'} eq "Suspended");
		return 1 if ($self->{'state'} eq "Target Full Duplex");
	}

	if ($self->{'rtype'} eq "GM") {
		return 0 if ($self->{'state'} eq "Copy Pending");
		return 0 if ($self->{'state'} eq "Suspended");
		return 1 if ($self->{'state'} eq "Target Copy Pending");
	}
	return 2;
}

sub exe_dscli {
	my $self = shift;
	my $cmdargs = shift;

	my $lpath = $self->{'limgid'} . "/" . $self->{'llss'};
	my $rpath = $self->{'rimgid'} . "/" . $self->{'rlss'};

	my @dscmd = @{$self->{'dscli'}};
	push(@dscmd, $cmdargs->{'cmd'});
	push(@dscmd, @{$cmdargs->{'flag'}})
		if (defined $cmdargs->{'flag'});
	push(@dscmd, "-dev", $self->{'limgid'});
	push(@dscmd, "-remotedev", $self->{'rimgid'})
		if (defined $cmdargs->{'rim'});

	push(@dscmd, @{$cmdargs->{'carg'}}) if (defined $cmdargs->{'carg'});

	if ($cmdargs->{'cmd'} eq "chsession") {
		push(@dscmd, "-lss", $self->{'llss'});
		push(@dscmd, "-action $cmdargs->{'act'}");
		my $volist = join(',', keys %{$self->{'vids'}});
		push(@dscmd, "-volume", $volist);
		push(@dscmd, $self->{'gmsess'});
	} elsif ($cmdargs->{'cmd'} =~ /fail(back|over)pprc/) {
		if ($self->{'rtype'} eq "MM") {
			push(@dscmd, "-type mmir");
		} elsif ($self->{'rtype'} eq "GM") {
			push(@dscmd, "-type gcp");
		} else {
			VCSAG_LOG_MSG("E", "Unsupported type $self->{'rtype'} for dscli command $cmdargs->{'cmd'}", 114, "$self->{'rtype'}", "$cmdargs->{'cmd'}");
			return 1;
		}

		push(@dscmd, "-tgtread")
			if ($cmdargs->{'cmd'} =~ /failbackpprc/);

		while (my ($lvid, $rvid) = each %{$self->{'vids'}}) {
			push(@dscmd, (join ':', $lvid, $rvid));
		}
	} elsif ($cmdargs->{'cmd'} =~ /(mk|resync|reverse)flash/) {
		push(@dscmd, @{$self->{'fl_pairs'}});
	} elsif ($cmdargs->{'cmd'} =~ /ls(pprc|flash)/) {
		push(@dscmd, keys %{$self->{'vids'}});
	} elsif ($cmdargs->{'cmd'} =~ /(mk|rm)gmir/) {
		push(@dscmd, "-lss", $self->{'llss'});
		if (defined $self->{'rmsess'}) {
			push(@dscmd, "-session", $self->{'rmsess'});
			push(@dscmd, join(":", $rpath, $lpath));
		} else {
			push(@dscmd, "-session", $self->{'gmsess'});
			push(@dscmd, join(":", $lpath, $rpath));
		}
	} elsif ($cmdargs->{'cmd'} =~ /mksession/) {
		push(@dscmd, "-lss", $self->{'llss'});
		push(@dscmd, $self->{'gmsess'});
	} elsif ($cmdargs->{'cmd'} =~ /showgmir/) {
		if (defined $cmdargs->{'getmast'}) {
			push(@dscmd, $rpath);
		} else {
			push(@dscmd, $self->{'llss'});
		}
	}

	my $ds_cmd = "@dscmd";
	VCSAG_LOG_MSG("N", "Executing dscli command: $ds_cmd", 115, "$ds_cmd")
		if (defined $cmdargs->{'pcmd'});
	$self->{'dscout'} = `@dscmd 2>&1`;
	my $ret = $?;
	my $dsc_out = $self->{'dscout'};
	VCSAG_LOG_MSG("E", "Execution of dscli command $ds_cmd returned: $dsc_out", 117, "$ds_cmd", "$dsc_out")
		if ($ret && (defined $cmdargs->{'perr'}));

	return $ret;
}

sub failback {
	my $self = shift;

	my $failbackpprc = {};
	$failbackpprc->{'cmd'} = "failbackpprc";
	$failbackpprc->{'rim'} = "";

	$failbackpprc->{'pcmd'} = "";
	$failbackpprc->{'perr'} = "";

	return ($self->exe_dscli($failbackpprc));
}

sub failover {
	my $self = shift;

	my $failoverpprc = {};
	$failoverpprc->{'cmd'} = "failoverpprc";
	$failoverpprc->{'rim'} = "";

	$failoverpprc->{'pcmd'} = "";
	$failoverpprc->{'perr'} = "";

	return ($self->exe_dscli($failoverpprc));
}

sub get_hostvol {
	#
	# Retrieve the device name as seen by the OS / vxvm.
	# This is used to enable diskgroups after a replication direction
	# switch or for fire drill 
	# - for mm / gm only for the MM / GC target volumes
	# - for mmsnap only for FC target volumes
	#
	my $self = shift;

	# rescan devices in case of AIX since MM / GM targte volumes
	# disappear during reboot
	if ($os_name =~ /AIX/i) {
		my $cfgmgr_err = `$aix_cfgmgr 2>&1`;
	}

	my @lshostvol = (@{$self->{'dscli'}}, "lshostvol", "-hdr off -bnr off");

	unless (defined(open(HVOL, "@lshostvol 2>&1 |"))) {
		VCSAG_LOG_MSG("E", "Error on open host volume command", 142);
		return 1;
	}

	while (<HVOL>) {
		my ($osdisk, $arvolid, undef) = split(/\s+/);
		my ($starid, $volid) = split('/',$arvolid);
		next unless ($starid =~ /$self->{'limgid'}/i);
		if ($self->{'rtype'} eq "MM" ||
		    $self->{'rtype'} eq "GM") {
			if (defined $self->{'vids'}->{"$volid"}) {
				$self->{'os_disk'}->{"$volid"} = $osdisk;
				$self->{'os_dlun'}->{"$osdisk"} = $volid;
				next;
			}
		} elsif ($self->{'rtype'} eq "MMS" ||
		    $self->{'rtype'} eq "GMS") {
			if (defined $self->{'fl_src'}->{"$volid"}) {
				$self->{'os_disk'}->{"$volid"} = $osdisk;
				$self->{'os_dlun'}->{"$osdisk"} = $volid;
			} elsif (defined $self->{'fl_tgt'}->{"$volid"}) {
				$self->{'os_disk'}->{"$volid"} = $osdisk;
				$self->{'os_dlun'}->{"$osdisk"} = $volid;
			}

		}
	}
	close HVOL;

	# Generate VXDISK information map
	if (-x $VXDG && -x $VXDISK) {
		my $vxdg_err = `$VXDCTL enable`;
		unless (defined(open(VXDSK, "$VXDISK -e list 2>&1 |"))) {
			VCSAG_LOG_MSG("E", "Error retrieving vxdisk information", 143);
		}
		while (<VXDSK>) {
			chomp;
			my ($enc_name, undef, undef, undef, $vx_stat, $nat_name) = split(/\s+/);
			next unless (defined $self->{'os_dlun'}->{"$nat_name"});
			if ($self->{'rtype'} eq "MMS" ||
			    $self->{'rtype'} eq "GMS") {
				my $skip_id = $self->{'os_dlun'}->{"$nat_name"};
				next unless (defined $self->{'fl_src'}->{"$skip_id"});
			}
			next unless ($vx_stat !~ /LVM/i);
			$self->{'vx_enabled'} = 1;
			$self->{'vx_name'}->{"$nat_name"} = $enc_name;
		}
		close VXDSK;
	}

	# Retrieve LVM information for AIX
	if ($os_name =~ /AIX/i) {
		unless (defined(open(LSPV, "$aix_lspv 2>&1 |"))) {
			VCSAG_LOG_MSG("E", "Error retrieving LVM information", 144);
		}
		while (<LSPV>) {
			chomp;
			my ($dsk_name, undef, $vg_name, undef) = split(/\s+/);
			next unless (defined $self->{'os_dlun'}->{"$dsk_name"});
			next if ($vg_name =~ /^VeritasVolumes$/i);
			next if ($vg_name =~ /^None$/i);
			$self->{'lvm_enabled'} = 1;
			$self->{'vg_name'}->{"$dsk_name"} = $vg_name;
		}
		close LSPV;
	}

	return 0;
}

# MetroMirror specific functions
# ==============================
#
sub get_mm_attr {
	my $self = shift;
	my @loc_attr = @_;

	# Get attributes
	#
	my $ret;
	my ($dscli_home, $hmc1, $hmc2, $user, $pwfile);
	my ($num_vids, $ind_vids, $tmp_vid);
	my $pathSeparator = "/";
	my $dscli_executable_name = "dscli";
    
	($ret, $dscli_home) = VCSAG_GET_ATTR_VALUE ("DSCliHome", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($dscli_home eq "")) {
		VCSAG_LOG_MSG("E", "No valid value for DSCliHome", 101);
		return 1;
	}

	($ret, $hmc1) = VCSAG_GET_ATTR_VALUE ("HMC1", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($hmc1 eq "")) {
		VCSAG_LOG_MSG("E", "No valid value for HMC1", 102);
		return 1;
	}

	($ret, $hmc2) = VCSAG_GET_ATTR_VALUE ("HMC2", -1, 1, @loc_attr);
	if ($ret != $VCSAG_SUCCESS) {
		VCSAG_LOG_MSG("E", "HMC2 is not part of the argument list", 103);
		return 1;
	}
	$hmc2 = '' if ($hmc2 eq '""');

	($ret, $user) = VCSAG_GET_ATTR_VALUE ("User", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($user eq "")) {
		VCSAG_LOG_MSG("E", "Invalid value for User", 104);
		return 1;
	}

	($ret, $pwfile) = VCSAG_GET_ATTR_VALUE ("PasswdFile", -1, 1, @loc_attr);
	$pwfile =~ s/^~/$ENV{'HOME'}/;
	unless (($ret == $VCSAG_SUCCESS) && (-r $pwfile)) {
		VCSAG_LOG_MSG("E", "Invalid value for PasswdFile", 105);
		return 1;
	}

	($ret, $self->{'limgid'}) = VCSAG_GET_ATTR_VALUE ("LocalStorageImageID", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($self->{'limgid'} eq "")) {
		VCSAG_LOG_MSG("E", "Invalid value for LocalStorageImageID", 106);
		return 1;
	}

	($ret, $self->{'rimgid'}) = VCSAG_GET_ATTR_VALUE ("RemoteStorageImageID", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($self->{'rimgid'} eq "")) {
		VCSAG_LOG_MSG("E", "Invalid value for RemoteStorageImageID", 107);
		return 1;
	}

	if ( $self->{'platform'} =~ /^Windows NT/i) {
		$dscli_home = Win32::GetShortPathName($dscli_home);
		$self->{'DSClihome'} = $dscli_home;
		$pwfile = Win32::GetShortPathName($pwfile);
		$pathSeparator = "\\";
		$dscli_executable_name =  $dscli_executable_name . ".exe";
	}


	$self->{'dscli'} = join $pathSeparator, $dscli_home, $dscli_executable_name;
	unless (-e $self->{'dscli'}) {
		VCSAG_LOG_MSG("E", "$self->{'dscli'} is not executable", 108, "$self->{'dscli'}");
		return 1;
	}

	if ($hmc2) {
		$self->{'dscli'} = [ $self->{'dscli'}, '-hmc1', $hmc1, '-hmc2', $hmc2, '-user', $user, '-pwfile', $pwfile ];
	} else {
		$self->{'dscli'} = [ $self->{'dscli'}, '-hmc1', $hmc1, '-user', $user, '-pwfile', $pwfile ];
	}

	# put volids in a hash where local volid is the key and
	# corresponding remote metromirror volid will be the value.
	# remote volid (i.e value field) to be populated later when determined.
	($ret, $num_vids, $ind_vids) = VCSAG_GET_ATTR_VALUE ("VolIds", @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($num_vids <= 0)) {
		VCSAG_LOG_MSG("E", "Number of VolIds is zero or invalid", 109);
		return 1;
	}

	for (my $ind = 0; $ind < $num_vids; $ind++) {
		($ret, $tmp_vid, $ind_vids) = VCSAG_GET_ATTR_VALUE ("VolIds", $ind_vids, ($ind + 1), @loc_attr);
		if ($ret != $VCSAG_SUCCESS || ($tmp_vid !~ /^[0-9a-f]+$/i)) {
			VCSAG_LOG_MSG("E", "Value $tmp_vid is not a valid hexadecimal volume ID", 110, "$tmp_vid");
			return 1;
		}
		$self->{'vids'}->{"$tmp_vid"} = "";
		$self->{'vids_state'}->{"$tmp_vid"} = "";
	}

	($ret, $self->{'DSMon'}) = VCSAG_GET_ATTR_VALUE ("DSMon", -1, 1, @loc_attr);
	$self->{'DSMon'} = 0 unless ($ret == $VCSAG_SUCCESS);

	($ret, $self->{'AutoFailback'}) = VCSAG_GET_ATTR_VALUE ("AutoFailback", -1, 1, @loc_attr);
	$self->{'AutoFailback'} = 1 unless ($ret == $VCSAG_SUCCESS);

	return 0;
}

# GlobalMirror specific functions
# ===============================
#
sub get_gm_attr {
	my $self = shift;
	my @loc_attr = @_;
	my $ret = 0;

	($ret, $self->{'llss'}) = VCSAG_GET_ATTR_VALUE ("LocalLSS", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || 
	    ($self->{'llss'} !~ /^[[:xdigit:]]{2}$/) ||
	    ($self->{'llss'} =~ /ff/i)) {
		VCSAG_LOG_MSG("E", "No valid resource value for LocalLSS, must be in range 00 - FE", 120);
		return 1;
	}

	($ret, $self->{'rlss'}) = VCSAG_GET_ATTR_VALUE ("RemoteLSS", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || 
	    ($self->{'rlss'} !~ /^[[:xdigit:]]{2}$/) ||
	    ($self->{'rlss'} =~ /ff/i)) {
		VCSAG_LOG_MSG("E", "No valid resource value for RemoteLSS, must be in range 00 - FE", 121);
		return 1;
	}

	($ret, $self->{'gmsess'}) = VCSAG_GET_ATTR_VALUE ("GMSession", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || 
	    ($self->{'gmsess'} !~ /^[[:xdigit:]]{2}$/)) {
		VCSAG_LOG_MSG("E", "No valid resource value for GMSession, must be in range 00 - FF", 122);
		return 1;
	}

    if ( $self->{'platform'} !~ /^Windows NT/i) {
        ($ret, $self->{'computedrsla'}) = VCSAG_GET_ATTR_VALUE ("ComputeDRSLA", -1, 1, @loc_attr);
        if (($ret != $VCSAG_SUCCESS) ||
            !(($self->{'computedrsla'} == 0) || ($self->{'computedrsla'} == 1)) ){
            VCSAG_LOG_MSG("E", "Not a valid value of ComputeDRSLA, must be either 0 or 1", 194);
            return 1;
        }
    }
	return 0;
}

sub gm_start_session {
	my $self = shift;

	# Check that all volumes have been added to the session
	my $mksess = {};
	$mksess->{'cmd'} = "mksession";

	if ($self->exe_dscli($mksess)) {
		unless ($self->{'dscout'} =~ /session already open/i) {
			VCSAG_LOG_MSG("E", "Global mirror session create command returned: $self->{'dscout'}", 123, "$self->{'dscout'}");
			return 1;
		}
	}

	my $chsess = {};
	$chsess->{'cmd'} = "chsession";
	$chsess->{'act'} = "add";
	$chsess->{'perr'} = "1";

	return 1 if ($self->exe_dscli($chsess));

	my $mkgmir = {};
	$mkgmir->{'cmd'} = "mkgmir";
	$mkgmir->{'perr'} = "1";

	return($self->exe_dscli($mkgmir));
}

sub gm_stop_session {
	my $self = shift;
	my $argv = shift;
	my $sub = 0;
	my $force = 0;

	if (defined $argv) {
		if ($argv =~ /sub/i) {
			$sub = 1;
		} elsif ($argv =~ /force/i) {
			$force = 1;
		}
	}

	if ($sub) {
		# Get the master session number
		my $showgmir = {};
		$showgmir->{'cmd'} = "showgmir";
		$showgmir->{'getmast'} = "";

		$self->exe_dscli($showgmir);
		foreach my $line (split('\n', $self->{'dscout'})) {
			next unless $line = s/^Master Session ID\s+0x//i;
			chomp($line);
			$self->{'rmsess'} = $line if ($line =~ /[0-9a-f]+/i);
			last;
		}
	}

	if ($sub == 0 || defined($self->{'rmsess'})) {
		my $rmgmir = {};
		$rmgmir->{'cmd'} = "rmgmir";
		push(@{$rmgmir->{'carg'}}, "-quiet");
		if (defined($self->{'rmsess'})) {
			push(@{$rmgmir->{'carg'}}, "-subordinate");
		} elsif ($force) {
			push(@{$rmgmir->{'carg'}}, "-force");
		}
		$rmgmir->{'perr'} = "1";

		$self->exe_dscli($rmgmir);
	}

	# remove volumes from session - otherwise failback may not work
	my $volist = join(',', keys %{$self->{'vids'}});

	my $chsess = {};
	$chsess->{'cmd'} = "chsession";
	$chsess->{'act'} = "remove";
	$chsess->{'perr'} = "1";

	return $self->exe_dscli($chsess);
}

sub gm_validate_flash {
	my $self = shift;

	if ($self->get_flash_vol()) {
		VCSAG_LOG_MSG("E", "Flash copy volume is a prerequisite for consistent global mirror failover", 129);
		return 1;
	}

	my $seqnum = 0;
	my $norevert = 0;
	my $commit = 0;
	foreach my $flvol (keys %{$self->{'vids'}}) {
		$seqnum = $self->{'seqnum'}->{"$flvol"} unless ($seqnum);
		if ($self->{'seqnum'}->{"$flvol"} ne $seqnum) { 
			VCSAG_LOG_MSG("E", "Flash copy volumes have inconsistent sequence numbers", 130);
			return 1;
		}
		$norevert = 1 if ($self->{'revert'}->{"$flvol"} =~ /Disable/i);
		$commit = 1 if ($self->{'revert'}->{"$flvol"} =~ /Enable/i);
	}

	if ($norevert) {
		return 0 unless ($commit);
		foreach my $flvol (keys %{$self->{'vids'}}) {
			next if ($self->{'revert'}->{"$flvol"} =~ /Disable/i);
			my $commflash = {};
			$commflash->{'cmd'} = "commitflash";
			$commflash->{'perr'} = "";
			push(@{$commflash->{'carg'}}, $flvol);

			return 1 if ($self->exe_dscli($commflash));
		}
	} else {
		foreach my $flvol (keys %{$self->{'vids'}}) {
			next if ($self->{'revert'}->{"$flvol"} =~ /Disable/i);
			my $revflash = {};
			$revflash->{'cmd'} = "revertflash";
			$revflash->{'perr'} = "";
			push(@{$revflash->{'carg'}}, $flvol);

			return 1 if ($self->exe_dscli($revflash));
		}
	}
	return 0;
}

sub reverse_flash {
	my $self = shift;

	my $reseflash = {};
	$reseflash->{'cmd'} = "reverseflash";
	push(@{$reseflash->{'carg'}}, "-fast -tgtpprc");
	$reseflash->{'perr'} = "1";

	return  1 if ($self->exe_dscli($reseflash));

	# Now wait until reverseflash has completed it's job
	my $rcnt = 1;
	while ($rcnt) {
		my $lsflash = {};
		$lsflash->{'cmd'} = "lsflash";
		push(@{$lsflash->{'flag'}}, "-bnr off -hdr off");
		$lsflash->{'perr'} = "";

		return  1 if ($self->exe_dscli($lsflash));

		last if ($self->{'cout'} =~ /lsflash:\s+No Flash Copy found/i);
		chomp $self->{'cout'};
		last if ($self->{'cout'} eq "");
		sleep($sltime);
	}

	my $mkflash = {};
	$mkflash->{'cmd'} = "mkflash";
	push(@{$mkflash->{'carg'}}, "-tgtinhibit -record -persist -nocp");
	$mkflash->{'perr'} = "";

	return $self->exe_dscli($mkflash);
}
	

# MetroMirrorSnap functions for fire drill
# ========================================
#
sub get_mms_attr {
	my $self = shift;
	my @loc_attr = @_;

	# Get attributes
	#
	my $ret;
	my ($target, $usesnap, $reqsnap, $msnap);

	($ret, $self->{'tres'}) = VCSAG_GET_ATTR_VALUE ("TargetResName", -1, 1, @loc_attr);
	if (($ret != $VCSAG_SUCCESS) || ($self->{'tres'} eq "")) {
		VCSAG_LOG_MSG("E", "No valid resource name for TargetResName", 136);
		return 1;
	}

        my ($type, undef) = split('\n', `$ag_i18n_inc::HARES -value $self->{'tres'} Type 2>&1`);
	unless ($type =~ /MetroMirror/i) {
		VCSAG_LOG_MSG("E", "Only resource type MetroMirror supported for $self->{'tres'}", 137, $self->{'tres'});
		return 1;
	}
	$self->{'targ'} = `$ag_i18n_inc::HARES -value $self->{'tres'} ArgListValues 2>&1`;

	if ($?) {
		VCSAG_LOG_MSG("E", "Can not get argument list for $self->{'tres'}", 138, $self->{'tres'});
		return 1;
	}

	($ret, $self->{'OffRefresh'}) = VCSAG_GET_ATTR_VALUE ("OffRefresh", -1, 1, @loc_attr);
	$self->{'OffRefresh'} = 0 unless ($ret == $VCSAG_SUCCESS);

	chomp($self->{'targ'});
	return 0;
}

sub get_flash_vol {
	my $self = shift;

	my $lsflash = {};
	$lsflash->{'cmd'} = "lsflash";
	push(@{$lsflash->{'flag'}}, "-fmt delim", "-delim ','");
	push(@{$lsflash->{'flag'}}, "-hdr off", "-bnr off -v off");
	$lsflash->{'perr'} = "1";

	return  1 if ($self->exe_dscli($lsflash));

	foreach my $line (split('\n', $self->{'dscout'})) {
		my($flpair, $srclss, $seqnum, undef, undef, 
			$record, undef, $revert, undef) = split(/,/, $line);
		my ($src_vol, $tar_vol) = split(/:/, $flpair);
		if (defined $self->{'vids'}->{"$src_vol"}) {
			# Store FC source as key, target as data
			$self->{'fl_tgt'}->{"$src_vol"} = $tar_vol;
			# Store FC target as key, source as data
			$self->{'fl_src'}->{"$tar_vol"} = $src_vol;
			# Store FC source as key, LSS as data
			$self->{'flss'}->{"$src_vol"} = $srclss;
			# Store FC source as key, FC seq number as data
			$self->{'seqnum'}->{"$src_vol"} = $seqnum;
			# Store FC source as key, FC revert state data
			$self->{'revert'}->{"$src_vol"} = $revert;
			if ($record =~ /enabled/i) {
				# save FC pairing as array
				push(@{$self->{'fl_pairs'}}, $flpair);
				$self->{'fl_recd'}->{"$src_vol"} = 1;
			}
		}
	}

	# Check if all replicated volumes have flash snap target
	# with the record option set, otherwise resyncflash will not work
	my $fl_error = 0;
	foreach my $srcvol (keys %{$self->{'vids'}}) {
		if (defined $self->{'fl_tgt'}->{"$srcvol"}) {
			unless (defined $self->{'fl_recd'}->{"$srcvol"}) {
				VCSAG_LOG_MSG("E", "Record flag not set for flash target of $srcvol, resync flash not possible", 140, $srcvol);
				$fl_error = 1;
			}
		} else {
			VCSAG_LOG_MSG("E", "No matching flash snap volume for $srcvol", 141, $srcvol);
			$fl_error = 1;
		}
	}
	return $fl_error;
}

sub refresh_snap {
	my $self = shift;

	my $resyflash = {};
	$resyflash->{'cmd'} = "resyncflash";
	push(@{$resyflash->{'carg'}}, "-record -persist");
	$resyflash->{'perr'} = "";

	return $self->exe_dscli($resyflash);
}

# Logical volume manager specific functions for replication & fire drill
# ======================================================================
#
sub vm_prep {
	my $self = shift;

	if ($self->{'rtype'} eq "MM" ||
	    $self->{'rtype'} eq "GM") {
		# enable replicated volumes if flash snap exists
		return 1 if ($self->get_hostvol());
		foreach my $vxdsk (keys %{$self->{'vx_name'}}) {
			my $vxupd = `$VXDISK updateudid $vxdsk 2>&1`;
		}
	} elsif ($self->{'rtype'} eq "MMS") {
		# enable flash snap volumes for FD
		return 1 if ($self->get_hostvol());

		my $lvm_err = $self->lvmvg_enable()
			if (defined $self->{'lvm_enabled'});
		my $vx_err = $self->vxdg_enable()
			if (defined $self->{'vx_enabled'});

	} elsif ($self->{'rtype'} eq "GMS") {
		# enable second flash for GM fire drill
	}
}

sub lvmvg_enable {
	my $self = shift;
	my %vgs;

	foreach my $vglun (keys %{$self->{'fl_tgt'}}) {
		my $mmdisk = $self->{'os_disk'}->{"$vglun"};
		my $vgdg = $self->{'vg_name'}->{"$mmdisk"};

		my $fc_lun = $self->{'fl_tgt'}->{"$vglun"};
		my $vgdisk = $self->{'os_disk'}->{"$fc_lun"};

		# extend volume group name for fire drill
		$vgdg .= $aix_fd_vg_ext;
		push(@{$vgs{"$vgdg"}}, $vgdisk);
	}

	foreach my $vgrp (keys %vgs) {
		my $exp_err = `$aix_exportvg $vgrp 2>&1`;
		my @cmd = ();
		push(@cmd, "$aix_recreatevg");
		push(@cmd, "-y $vgrp");
		push(@cmd, "-Y $aix_lv_prefix");
		push(@cmd, "-L $aix_label_prefix");
		push(@cmd, @{$vgs{"$vgrp"}});

		my $rec_mess = `@cmd 2>&1`;
		if ($?) {
			VCSAG_LOG_MSG("E", "Error when recreating volume group: $rec_mess", 146, $rec_mess);
			return 1;
		}

	}

	# Find dependent groups for major number update
	open(RDEP, "$ag_i18n_inc::HARES -dep $self->{'resname'} 2>&1 |");

	while (<RDEP>) {
		next if /^#/;
		chomp;
		my (undef, $parent, $child) = split(/\s+/);
		next unless ($child eq $self->{'resname'});
		my $pa_type = `$ag_i18n_inc::HARES -value $parent Type 2>&1`;
		next unless ($pa_type =~ /LVMVG/);
		my $pa_vg = `$ag_i18n_inc::HARES -value $parent VolumeGroup 2>&1`;
		chomp($pa_vg);
		# Need to update major number
		if (defined $vgs{"$pa_vg"}) {
			my $mj_err = ArrayRep::aix_lvm_update_majornum_attrib($parent, \@{$vgs{"$pa_vg"}});
		}
	}
	close RDEP;
	return 0;
}

sub vxdg_enable {
	my $self = shift;

	my $dgnamelist = "";
	my $tagname = $self->{'resname'};
	foreach my $vxdsk (keys %{$self->{'vx_name'}}) {
		# Get current diskgroup name
		my $dname = $self->{'vx_name'}->{"$vxdsk"};
		open(VXDK, "$VXDISK list $dname 2>&1 |");
		while (<VXDK>) {
			next unless /^group:/;
			my (undef, $vgrp, undef) = split(/\s+/);
			$vgrp =~ s/^name=//;
			return 1 if
			   (ArrayRep::vxvm_prepare_disks($tagname, $is_rac_cluster, $dname));
			if ($dgnamelist eq ""){
				$dgnamelist = "$vgrp";
			} elsif ($dgnamelist !~ /\b$vgrp\b/) {
				$dgnamelist .= " " . "$vgrp";
			}
			last;
		}
		close VXDK;
	}
	return 1 if ($dgnamelist eq "");
	return 1 if (ArrayRep::vxvm_test_diskgroup_import($dgnamelist, "gold", $tagname, $is_rac_cluster));
	return 0;
}

# VMware ESX specific extensions to library
# =========================================
#
sub get_con_str {
	my ($vprovider, $host, $username, $password, $keystore, $code);
	open(CRED, "<$cred_file" || die "Cannot open $cred_file for reading: $!");
	my @lines = <CRED>;
	close (CRED);
	$username = "root";
	my $localhost = `uname -n`;
	chomp($localhost);
	my @vcshostname = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
	my $vcshost = $vcshostname[5];
	foreach (@lines) {
		if($_ =~ /$vcshost/) {
			($host, $code, $password, $keystore) = split('\s+');
			last;
		}
	}

	#for backwards compatibility with VCS 5.1 GA
	if ($code eq "0") {
		$username = "root";
	} elsif ($code eq "1") { 
		$username = "vcs";
	} else {
		$username = $code;
	}

	$vprovider = "1";
	my $dpassword=`$VCSHOME/sbin/vcsdecrypt $password`;
	$host = "https://" . $localhost . "/sdk";
	my @ops = ($vprovider, $host, $username, $dpassword, $keystore);
	my $con_str= join(' ', @ops);
	return $con_str;
}

sub check_vmware_api_err {
	my $localhost = `uname -n`;
	chomp($localhost);
	my $op = $_[0];
	my $errcode = $_[1];
	my $vmtask = $_[2];
	if($errcode == 1) {
		VCSAG_LOG_MSG("E", "VI3 API could not login to $localhost\n", 58, $localhost);
		VCSAG_LOG_MSG("E", "It could either be a network config issue (DNS failure) or invalid localhost credentials file", 59);
	} elsif ($errcode == 2) {
		VCSAG_LOG_MSG("E", "$vmtask failed: $localhost\n", 57, $vmtask, $localhost);
	}
}

# rescan hba and vmfs for new datastores
sub exec_vmware_api_cmd {
	my $operation = shift;
	my $localhost = `uname -n`;
	chomp($localhost);
	my $con_str = get_con_str();
	my @ops = ($con_str, $operation, $localhost);
	my $operstr = join(' ', @ops);
	my $op = `$VMWARE_API_CMD $operstr`;
	my $retcode = $? >> 8;
	chomp($op);
	check_vmware_api_err($localhost, $op, $retcode, $operation);
	return 0;
}

# Check if vml device entries already exist
sub vml_check {
	my($self) = shift;
	# Check if disk entries for the volumes do not already exist
	my %Vols;
	my $tmpHBA;

	# Get the WWN of the array
	my @lssu = (@{$self->{dscli}}, "lssu", "-fmt delim", "-delim ','", "-hdr off", "-bnr off", "-v off", $self->{'limgid'});
	my $WWNcmp;
	$_ = `@lssu`;
	(undef, undef, undef, $WWNcmp, undef) = split /,/;
	# Strip off first number since it differs
	$WWNcmp =~ s/^.//;

	# Get the list of vml entries for this array
	open (VMHBAD, "esxcfg-mpath -l -v |");
	while(<VMHBAD>) {
		# Extract the UUID as reported by GetVMDiskUUID call
		next unless /^Disk/;
		(undef, undef, $tmpHBA, undef) = split;
		$tmpHBA =~ tr/a-f/A-F/;
		next unless $tmpHBA =~ s/^vml\..{11}$WWNcmp(.{16}).*/$1/;	
		$tmpHBA =~ s/^0+//;
		$Vols{"$tmpHBA"} = 1;
	}
	close VMHBAD;
	if (keys %Vols) {
		while (my ($lvid, $rvid) = each %{$self->{'vids'}}) {
			next if defined $Vols{"$lvid"};
			return 1;
		}
	} else {
		return 1;
	}
	return 0;
}

sub vmware_online {
	my($self) = shift;

	my $cmd = "";

	# Supress rescan if device entries already exist
	unless ($self->vml_check()) {
		exec_vmware_api_cmd("RefreshHostStorage");
	} else {
		# Make sure that /proc/vmware/config/LVM/EnableResignature and
		# /proc/vmware/config/LVM/DisallowSnapshotLun both are set to 0
		$cmd = "esxcfg-advcfg -s 0 /LVM/EnableResignature";
		system ($cmd);
		$cmd = "esxcfg-advcfg -s 0 /LVM/DisallowSnapshotLun";
		system ($cmd);

		VCSAG_SET_RES_EP_TIMEOUT(180);
		exec_vmware_api_cmd("RescanAllHba");
	}
	return 0;
}

1;
